program triangular
  implicit none
! a is the data structure used in this program; a_serial was used
! in the serial program
! a_serial(i)  =  a(i/nprocs)[mod(i,nprocs)]  
!                                             i = 1..n
! a_serial(me+(i_local-1)*nprocs) = a(i_local)[me]  
!                                             i_local = 1..rows_per_proc
!                                             me = 1..nprocs 
!
  type :: tri_matrix
    real, allocatable :: row(:)
  end type
  type(tri_matrix), allocatable :: a(:)[:]

  integer :: n[*], i_row[*] 
  integer ::  nproc, me, rows_per_proc
  integer ::  i, j, i_local, n_elem

  nproc = num_images()
  me = this_image()
  if (me == 1) then
     write (*, fmt='(''number of rows (1..100):'')')
     read(*, *) n 
     write (*, fmt='(''which row should be printed (1..'',i4,''):'')') n 
     read(*, *) i_row
  end if   
  sync all           ! Now, n[1] and i_row[1] are accessible from each image.
  n = n[1]           ! Copying of information into the local portions of
  i_row = i_row[1]   ! variables n and i_row on each image.
  
  rows_per_proc = (n-1) / nproc + 1
  allocate(a(rows_per_proc)[*])
!
! create triangular matrix :
  i_local = 1
  n_elem = 0
  do i = me, n, nproc
    allocate(a(i_local)%row(n - i + 1))
    n_elem = n_elem + n - i + 1
    i_local = i_local + 1
  end do
!
! initialize matrix A(i, j) = i + j  
  i_local = 1
  do i = me, n, nproc
    do j = 1, n - i + 1
       a(i_local)%row(j) = real(i) + real(j)
    end do
    i_local = i_local + 1
  end do
!
! write row number i_row to standard output
  if (mod(i_row - 1, nproc) + 1 == me .and. i_row <= n) then
    i_local = (i_row - 1) / nproc + 1
    write(*, fmt='((''[me='',i4'']  Row '',i4,'' is local on image '',i4,'', i_local='',i4,'':''),1000(1X,F4.0))') &
          me, i_row, mod(i_row - 1, nproc) + 1, i_local, &
          (a(i_local)%row(j),j=1, n - i_row + 1)
  end if

  write(*, fmt='(''Number of elements on image '',i0,'': '',i0)') &
           me, n_elem
!  
! clean up
  sync all
! the following three lines should not be needed
! (workaround for a bug in some compilers)
  do i=1, size(a)
    if (allocated(a(i)%row)) deallocate(a(i)%row)
  end do
  sync all
  deallocate(a)  ! implies deallocation of components
end program
