#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h> 

#define min(A,B) ((A) < (B) ? (A) : (B))
#define max(A,B) ((A) > (B) ? (A) : (B))

#ifndef prtrows
#  define prtrows 11
#endif 
#ifndef prtcols
#  define prtcols 11
#endif 
 
int m,n,itmax; 
#define idx(i,k) ((i)*n+(k)) 
double *phi; /* usage: phi[m*n] == phi[m][n], phi[idx(i,k)] = phi[i][k] */
double *phin; 
 
void heatpr();

void heat(); 

int main(int argc, char **argv)
{
  int dim1, dim2;
  if (argc < 3) 
  {
      printf("Synopsis: heat <dim1> <dim2> [<itmax>]\n");
      printf("\n");
      printf("Examples: heat 30 30\n");
      printf("          heat 30 30 15000\n");
    exit(1); 
  } 
  dim1 = atoi(argv[1]); 
  dim2 = atoi(argv[2]); 
  itmax=15000; 
  if (argc > 3) itmax = atoi(argv[3]); 
 
  m=dim1+2; n=dim2+2; /* two additional elements for the boundary conditions */ 
  printf("main:   dim1  =%4d,   dim2  =%4d\n", dim1, dim2);
  printf("main: m=dim1+2=%4d, n=dim2+2=%4d\n", m, n);
  phi =(double *)malloc(m*n*sizeof(double)); 
  phin=(double *)malloc(m*n*sizeof(double)); 
  heat();
  exit(0); 
}

void heat() 
{
  double eps = 1.0e-08;
  double dx,dy,dx2,dy2,dx2i,dy2i,dt,dphi,dphimax;
  int i,k,it;

  clock_t t1,t2;
  struct timeval tv1,tv2; struct timezone tz;
 
  dx=1.0/(n-1);
  dy=1.0/(m-1);
  dx2=dx*dx;
  dy2=dy*dy;
  dx2i=1.0/dx2;
  dy2i=1.0/dy2;
  dt=min(dx2,dy2)/4.0;
/* start values - total area: 0.d0 */
  for (i=0;i<m;i++)
  {
    for (k=0;k<n;k++)
    {
      phi[idx(i,k)]=0.0;
    }
  }
/* start values - boundary: 1.d0 */
  for (i=0;i<m;i++)
  {
    phi[idx(i,n-1)]=1.0;
  }
/* start values - boundary: 0.d0 */
  for (i=0;i<m;i++)
  {
    phi[idx(i,0)]=0.0;
  }
/* start values - boundary: continuously from 0 to 1 */
  for (k=1;k<n-1;k++)
  {
    phi[idx(0,k)]=k*dx;
    phi[idx(m-1,k)]=k*dx;
  }

  printf("\nHeat Conduction 2d\n");
  printf("\ndx = %12.4g, dy = %12.4g, dt = %12.4g, eps = %12.4g\n",
         dx,dy,dt,eps);
  printf("\nstart values\n");
  heatpr(m, n, phi);

  gettimeofday(&tv1, &tz);
  t1=clock();

/* iteration */
  for (it=1;it<=itmax;it++)
  {
    dphimax=0.;
    for (i=1;i<m-1;i++)
    {
      for (k=1;k<n-1;k++)
      {
        dphi=(phi[idx(i+1,k)]+phi[idx(i-1,k)]-2.*phi[idx(i,k)])*dy2i
            +(phi[idx(i,k+1)]+phi[idx(i,k-1)]-2.*phi[idx(i,k)])*dx2i;
        dphi=dphi*dt;
        dphimax=max(dphimax,dphi);
        phin[idx(i,k)]=phi[idx(i,k)]+dphi;
      }
    }
/* save values */
    for (i=1;i<m-1;i++)
    {
      for (k=1;k<n-1;k++)
      {
        phi[idx(i,k)]=phin[idx(i,k)];
      }
    }
    if(dphimax<eps) break;
  }
  if(dphimax>=eps) it=it-1;

  t2=clock();
  gettimeofday(&tv2, &tz);

  printf("\nphi after %d iterations, dphimax=%9.3e, eps=%9.3e\n",it, dphimax, eps);
  heatpr(m, n, phi);
  printf( "CPU time (clock)                = %12.6f sec\n", (t2-t1)/1000000.0 );
  printf( "wall clock time (gettimeofday)  = %12.6f sec\n", (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6 );

}

void heatpr()
{
  int i,k, prt_i,prt_k, prt_i_max, prt_k_max;

    prt_i_max = min(prtrows, m); 
    prt_k_max = min(prtcols, n); 
    printf("i  \\  k="); 
    for (prt_k=0; prt_k<prt_k_max; prt_k++)
    { k=1.0*prt_k/(prt_k_max-1)*(n-1) + 0.5/*rounding*/;
      printf("  %6i",k); 
    } 
    printf("\n"); 
    printf("-------+"); 
    for (prt_k=0; prt_k<prt_k_max; prt_k++)
    { k=1.0*prt_k/(prt_k_max-1)*(n-1) + 0.5/*rounding*/;
      printf("--------"); 
    } 
    printf("\n"); 
    for (prt_i=0; prt_i<prt_i_max; prt_i++)
    { i=1.0*prt_i/(prt_i_max-1)*(m-1) + 0.5/*rounding*/;
      printf("%6i |",i); 
      for (prt_k=0; prt_k<prt_k_max; prt_k++)
      { k=1.0*prt_k/(prt_k_max-1)*(n-1) + 0.5/*rounding*/;
        printf("  %6.4f",phi[idx(i,k)]);
      }
      printf("\n");
    }

return;
}
