#include <stdio.h>
#include <stdlib.h>
#include <upc.h>
//
// A is addressed in the same manner as in the serial case
// 
typedef struct {
  shared [] float *row;
  size_t row_size;
} Tri_matrix;

shared [1] size_t n, i_row;
shared [1] Tri_matrix *A;

int main(void) {
  size_t i, i_local, j, n_elem;
  int next_thread;
  float *row;

  if (MYTHREAD == 0) {
    printf("number of rows (1..100):\n");
    scanf("%li",(size_t *) &n);
    printf("which row should be printed (1..%1li):\n",n);
    scanf("%li",(size_t *) &i_row);
//   casts to (size_t *) are required and work because data are located
//   on executing thread
  }
  upc_barrier;
// n and i_row are now accessible from each thread

  A = (shared [1] Tri_matrix *) 
       upc_all_alloc(n, sizeof(Tri_matrix));
//
// create triangular matrix
  n_elem = 0;
  for (i=MYTHREAD; i<n; i+=THREADS) {
    A[i].row_size = n - i;
    A[i].row = 
         (shared [] float *) upc_alloc(A[i].row_size*sizeof(float));
    n_elem += n - i;
  }
//
// initialize matrix A(i, j) = i + j
  for (i=MYTHREAD; i<n; i+=THREADS) {
//  cast to local pointer for performance
    row = (float *) A[i].row;
    for (j=0; j<n-i; j++) {
      row[j] = (float) (i + 1) + (float) (j + 1);
    }
  }
//
// write (one-based) row number i_row to standard output
  if ( (i_row - 1)%THREADS == MYTHREAD && i_row >= 1 && i_row <= n ) {
    printf("[me=%2i]  Row %li on thread %i:",  MYTHREAD, i_row, (i_row - 1)%THREADS);
//  printf(" (size=%2ld=%2ld)", A[      i_row-1 ].row_size,n-i_row+1);
    printf(" (size=%2ld=%2ld)", A[(int)(i_row-1)].row_size,n-i_row+1); // the cast (int) is only necessary due to a complier bug
    for (j=0; j<n-i_row+1; j++) {
//    printf(" %4.0f", A[      i_row-1 ].row[j]);
      printf(" %4.0f", A[(int)(i_row-1)].row[j]); // the cast (int) is only necessary due to a complier bug
    }
    printf("\n");
  }
  printf("Number of elements on thread %i: %li\n", 
          MYTHREAD,n_elem);
//
// clean up
  upc_barrier; 
  for (i=MYTHREAD; i<n; i+=THREADS) {
    upc_free(A[i].row);
  }
  upc_barrier; 
  if (MYTHREAD == 0) upc_free(A);
}
